/*
  Copyright 2021 Kakao Corp.

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
 */
package com.kakao.sdk.partner.friend.adapter

import android.content.res.Configuration
import android.util.TypedValue
import android.view.LayoutInflater
import android.view.ViewGroup
import android.widget.LinearLayout
import androidx.constraintlayout.widget.ConstraintSet
import androidx.constraintlayout.widget.ConstraintSet.*
import androidx.recyclerview.widget.DiffUtil
import androidx.recyclerview.widget.RecyclerView
import com.kakao.sdk.partner.friend.imageloader.ImageLoader
import com.kakao.sdk.partner.friend.model.Item
import com.kakao.sdk.partner.friend.util.LayoutUtils
import com.kakao.sdk.partner.friend.util.dpToPixel
import com.kakao.sdk.v2.partner.friend.R
import com.kakao.sdk.v2.partner.friend.databinding.CategoryDividerBinding
import com.kakao.sdk.v2.partner.friend.databinding.EmptySearchResultTextBinding
import com.kakao.sdk.v2.partner.friend.databinding.HeaderFriendPickerBinding
import com.kakao.sdk.v2.partner.friend.databinding.ItemFriendPickerBinding
import kotlin.math.abs

class FriendsListAdapter(
    private val isPopup: Boolean,
    private val isSingle: Boolean,
    private val maxPickableCount: Int,
    private val checkCallback: (Item.User?) -> Unit,
    private val uncheckCallback: (Item.User) -> Unit
) : RecyclerView.Adapter<RecyclerView.ViewHolder>() {
    private var items = mutableListOf<Item>() // 화면에 보여지는 전체 아이템 리스트
    private var pickedFriends = mutableListOf<Item.User>() // 친구 추가 선택 가능 여부 파악을 위한 리스트

    override fun onCreateViewHolder(parent: ViewGroup, viewType: Int): RecyclerView.ViewHolder {
        val inflater = LayoutInflater.from(parent.context)
        return when (viewType) {
            VIEW_TYPE_HEADER -> {
                HeaderViewHolder(
                    isPopup,
                    HeaderFriendPickerBinding.inflate(inflater, parent, false)
                )
            }
            VIEW_TYPE_USER -> {
                UserViewHolder(isPopup,
                    ItemFriendPickerBinding.inflate(inflater, parent, false),
                    checkCallback = {
                        // 멀티 피커이면서 이미 최대 선택 가능한 친구를 다 채운 경우 null 리턴
                        if (pickedFriends.size == maxPickableCount) {
                            // 싱글 피커인 경우 이전에 선택된 친구 삭제 후 새로운 친구 리턴
                            if (isSingle) {
                                uncheckCallback(pickedFriends[0])
                                checkCallback(it)
                            } else {
                                checkCallback(null)
                            }
                        } else {
                            checkCallback(it)
                        }
                    },
                    uncheckCallback = {
                        uncheckCallback(it)
                    })
            }
            VIEW_CATEGORY_DIVIDER -> {
                DividerViewHolder(CategoryDividerBinding.inflate(inflater, parent, false))
            }
            VIEW_EMPTY_SEARCH_RESULT_TEXT -> {
                EmptySearchViewHolder(
                    isPopup,
                    EmptySearchResultTextBinding.inflate(inflater, parent, false)
                )
            }
            else -> throw ClassCastException("Unknown viewType $viewType")
        }
    }

    override fun onBindViewHolder(holder: RecyclerView.ViewHolder, position: Int) {
        val item = items[position]
        when (holder) {
            is HeaderViewHolder -> holder.bind(item as Item.Header)
            is UserViewHolder -> holder.bind(item as Item.User, isSingle)
            is EmptySearchViewHolder -> holder.setLayout()
        }
    }

    override fun getItemCount(): Int = items.size

    override fun getItemViewType(position: Int): Int {
        return when (items[position]) {
            is Item.Header -> VIEW_TYPE_HEADER
            is Item.User -> VIEW_TYPE_USER
            is Item.Divider -> VIEW_CATEGORY_DIVIDER
            is Item.EmptySearchResult -> VIEW_EMPTY_SEARCH_RESULT_TEXT
        }
    }

    fun setFriendsList(items: MutableList<Item>) {
        val diffUtilCallback = ItemDiffUtilCallback(this.items, items)
        val diffResult = DiffUtil.calculateDiff(diffUtilCallback)
        diffResult.dispatchUpdatesTo(this)
        this.items = items
    }

    fun checkFriend(friend: Item.User) {
        updateDuplicatedFriend(friend, true)
    }

    fun uncheckFriend(friend: Item.User) {
        updateDuplicatedFriend(friend, false)
    }

    fun setPickedFriends(pickedFriends: MutableList<Item.User>) {
        this.pickedFriends = pickedFriends
    }

    private fun updateDuplicatedFriend(user: Item.User, isChecked: Boolean) {
        val indexOfFirst = items.indexOfFirst { it == user }
        val indexOfLast = items.indexOfLast { it == user }
        // index를 못찾은 경우에는 -1 이므로 crash 방지 (검색 중에는 리스트에 해당 user가 없을 수 있음)
        if (indexOfFirst > 0) {
            (items[indexOfFirst] as Item.User).isChecked = isChecked
        }
        if (indexOfLast > 0) {
            (items[indexOfLast] as Item.User).isChecked = isChecked
        }
        notifyItemChanged(indexOfFirst)
        notifyItemChanged(indexOfLast)
    }

    class HeaderViewHolder(
        private val isPopup: Boolean,
        private val binding: HeaderFriendPickerBinding
    ) : RecyclerView.ViewHolder(binding.root) {
        fun bind(header: Item.Header) {
            if (isPopup) {
                setPopupLayout()
            }
            binding.titleTv.text = header.title
        }

        private fun setPopupLayout() {
            binding.apply {
                titleTv.apply {
                    ConstraintSet().apply {
                        clone(root)
                        connect(id, TOP, PARENT_ID, TOP, dpToPixel(10f, resources))
                        connect(id, BOTTOM, PARENT_ID, BOTTOM, dpToPixel(6f, resources))
                    }.applyTo(root)
                }
            }
        }
    }

    class DividerViewHolder(binding: CategoryDividerBinding) : RecyclerView.ViewHolder(binding.root)
    class EmptySearchViewHolder(
        private val isPopup: Boolean,
        private val binding: EmptySearchResultTextBinding
    ) : RecyclerView.ViewHolder(binding.root) {
        fun setLayout() {
            binding.apply {
                if (LayoutUtils.orientation == Configuration.ORIENTATION_PORTRAIT) {
                    if (isPopup) {
                        emptySearchResultText.apply {
                            layoutParams = (layoutParams as LinearLayout.LayoutParams).apply {
                                topMargin = dpToPixel(60f, resources)
                            }
                            setTextSize(TypedValue.COMPLEX_UNIT_SP, 14f)
                        }
                    }
                } else {
                    emptySearchResultText.apply {
                        layoutParams = (layoutParams as LinearLayout.LayoutParams).apply {
                            topMargin = if (isPopup) {
                                dpToPixel(20f, resources)
                            } else {
                                dpToPixel(30f, resources)
                            }
                        }
                        if (isPopup) {
                            setTextSize(TypedValue.COMPLEX_UNIT_SP, 14f)
                        }
                    }
                }
            }
        }
    }

    class UserViewHolder(
        private val isPopup: Boolean,
        private val binding: ItemFriendPickerBinding,
        private val checkCallback: (Item.User) -> Unit,
        private val uncheckCallback: (Item.User) -> Unit
    ) : RecyclerView.ViewHolder(binding.root) {

        fun bind(user: Item.User, isSingle: Boolean) {
            binding.apply {
                if (isPopup) {
                    setPopupLayout()
                }
                val value = TypedValue()
                when (abs(user.uuid?.hashCode()?.rem(3) ?: 0)) {
                    0 -> root.context.theme.resolveAttribute(
                        R.attr.default_profile_image1,
                        value,
                        true
                    )
                    1 -> root.context.theme.resolveAttribute(
                        R.attr.default_profile_image2,
                        value,
                        true
                    )
                    2 -> root.context.theme.resolveAttribute(
                        R.attr.default_profile_image3,
                        value,
                        true
                    )
                }
                ImageLoader.get(itemView.context).load(user.profileImageUrl)
                    .placeHolder(value.resourceId).into(profileIv)
                nicknameTv.text = user.nickname
                if (isSingle) {
                    checkBox.setButtonDrawable(R.drawable.daynight_radio)
                } else {
                    checkBox.setButtonDrawable(R.drawable.daynight_friends_picker_checkbox)
                }
                checkBox.isChecked = user.isChecked

                user.uuid?.let {
                    root.setOnClickListener {
                        if (!checkBox.isChecked) {
                            checkCallback(user)
                        } else {
                            uncheckCallback(user)
                        }
                    }
                }
            }
        }

        private fun setPopupLayout() {
            binding.apply {
                val size = dpToPixel(30f, root.resources)
                profileIv.apply {
                    layoutParams.apply {
                        width = size
                        height = size
                    }
                    ConstraintSet().apply {
                        clone(root)
                        connect(id, TOP, PARENT_ID, TOP, dpToPixel(7f, resources))
                        connect(id, BOTTOM, PARENT_ID, BOTTOM, dpToPixel(7f, resources))
                    }.applyTo(root)
                }
                nicknameTv.apply {
                    ConstraintSet().apply {
                        clone(root)
                        connect(id, START, profileIv.id, END, dpToPixel(8f, resources))
                        connect(id, TOP, PARENT_ID, TOP, dpToPixel(11f, resources))
                        connect(id, BOTTOM, PARENT_ID, BOTTOM, dpToPixel(12.5f, resources))
                    }.applyTo(root)
                    setTextSize(TypedValue.COMPLEX_UNIT_SP, 14f)
                }
                checkBox.apply {
                    ConstraintSet().apply {
                        clone(root)
                        connect(id, TOP, PARENT_ID, TOP, dpToPixel(13f, resources))
                        connect(id, BOTTOM, PARENT_ID, BOTTOM, dpToPixel(13f, resources))
                    }.applyTo(root)
                }
            }
        }
    }

    companion object {
        const val VIEW_TYPE_HEADER = 0
        const val VIEW_TYPE_USER = 1
        const val VIEW_CATEGORY_DIVIDER = 2
        const val VIEW_EMPTY_SEARCH_RESULT_TEXT = 3
    }
}

class ItemDiffUtilCallback(private val oldList: List<Item>, private val newList: List<Item>) :
    DiffUtil.Callback() {
    override fun getOldListSize(): Int = oldList.size

    override fun getNewListSize(): Int = newList.size

    override fun areItemsTheSame(oldItemPosition: Int, newItemPosition: Int): Boolean {
        return oldList[oldItemPosition].hashCode() == newList[newItemPosition].hashCode()
    }

    override fun areContentsTheSame(oldItemPosition: Int, newItemPosition: Int): Boolean {
        val oldItem = oldList[oldItemPosition]
        val newItem = newList[newItemPosition]
        when (oldItem) {
            is Item.Header -> {
                if (newItem !is Item.Header) {
                    return false
                }
                return oldItem.title == newItem.title
            }
            is Item.User -> {
                if (newItem !is Item.User) {
                    return false
                }
                return oldItem.uuid == newItem.uuid
            }
            is Item.Divider -> {
                if (newItem !is Item.Divider) {
                    return false
                }
                return true
            }
            is Item.EmptySearchResult -> {
                if (newItem !is Item.EmptySearchResult) {
                    return false
                }
                return true
            }
        }
    }
}