/*
  Copyright 2021 Kakao Corp.

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
 */
package com.kakao.sdk.partner.friend.viewmodel

import android.app.Application
import android.os.Parcelable
import android.os.ResultReceiver
import androidx.lifecycle.*
import com.kakao.sdk.partner.friend.Repository
import com.kakao.sdk.partner.friend.model.Item
import com.kakao.sdk.partner.friend.model.OsFilter
import com.kakao.sdk.partner.friend.model.RequestPickerParams
import com.kakao.sdk.partner.friend.util.FriendComparator
import com.kakao.sdk.partner.friend.util.PhonemeUtils
import com.kakao.sdk.partner.talk.model.PartnerFriendsContext
import com.kakao.sdk.v2.partner.friend.R
import java.util.*
import kotlin.properties.Delegates

class FriendViewModelFactory(
    private val application: Application,
    private val resultReceiver: ResultReceiver,
    private val param: Parcelable,
    private val isSingle: Boolean,
    private val isPopup: Boolean
) : ViewModelProvider.AndroidViewModelFactory(application) {
    override fun <T : ViewModel?> create(modelClass: Class<T>): T {
        return if (modelClass.isAssignableFrom(FriendViewModel::class.java)) {
            FriendViewModel(application, resultReceiver, param, isSingle, isPopup) as T
        } else {
            throw IllegalArgumentException()
        }
    }
}

class FriendViewModel(
    application: Application,
    val resultReceiver: ResultReceiver,
    val param: Parcelable,
    val isSingle: Boolean,
    val isPopup: Boolean
) : AndroidViewModel(application) {
    private val _pickerList = MutableLiveData<MutableList<Item>>()
    val pickerList: LiveData<MutableList<Item>> get() = _pickerList

    private val _pickedFriends = MutableLiveData<MutableList<Item.User>>()
    val pickedFriends: LiveData<MutableList<Item.User>> get() = _pickedFriends

    private val totalList = mutableListOf<Item>()

    private val _friendsList = mutableListOf<Item.User>() // 불러온 모든 친구 데이터가 저장된 리스트
    val friendsList: MutableList<Item.User> get() = _friendsList

    private val _favoritesList = mutableListOf<Item.User>()

    private val _pickedFriendsList = mutableListOf<Item.User>()
    val pickedFriendsList: MutableList<Item.User> get() = _pickedFriendsList.toMutableList()

    private val _isSearching = MutableLiveData<Boolean>()
    val isSearching: LiveData<Boolean> get() = _isSearching

    private val _isError = MutableLiveData<Boolean>()
    val isError: LiveData<Boolean> get() = _isError

    private val repository = Repository

    private var friendsContext: PartnerFriendsContext? = null

    private var showMyProfile by Delegates.notNull<Boolean>()
    private var showFavorite by Delegates.notNull<Boolean>()
    private var osFilter by Delegates.notNull<OsFilter>()

    private val comparator: FriendComparator = FriendComparator() // 친구 리스트 정렬용 Comparator

    private val context = application.applicationContext

    init {
        (param as RequestPickerParams).let {
            showMyProfile = it.showMyProfile!!
            showFavorite = it.showFavorite!!
            osFilter = it.usingOsFilter!!
        }
    }

    fun loadData() {
        repository.friends
            .filter { if (osFilter != OsFilter.ALL) it.talkOs == osFilter.name.toLowerCase(Locale.getDefault()) else true } // os 필터링
            .forEach { user ->
                val friend = Item.User(user.profileNickname, user.profileThumbnailImage, user.uuid)
                if (user.favorite) {
                    _favoritesList.add(friend)
                }
                _friendsList.add(friend)
            }
        _friendsList.sortWith(comparator)

        // 친구가 한 명도 없는 경우
        if (_friendsList.size == 0) {
            _pickerList.value = totalList
            _isError.value = true
            return
        }

        if (showMyProfile) {
            totalList.add(Item.Header(context.getString(R.string.my_profile)))
            repository.me?.let {
                totalList.add(
                    Item.User(
                        it.profileNickname,
                        it.profileThumbnailImage,
                        it.uuid,
                        isMyProfile = true
                    )
                )
            }
            totalList.add(Item.Divider)
        }
        if (showFavorite && _favoritesList.size > 0) {
            totalList.add(Item.Header(context.getString(R.string.favorite)))
            totalList.addAll(_favoritesList)
            totalList.add(Item.Divider)
        }
        totalList.add(Item.Header(context.getString(R.string.friend)))
        totalList.addAll(_friendsList)
        _pickerList.value = totalList
    }

    fun retryLoadFriendsData() {
        friendsContext = null
        totalList.clear()
        loadData()
    }

    fun addCheckedFriend(user: Item.User) {
        _pickedFriendsList.add(0, user)
        _pickedFriends.value = _pickedFriendsList
        // 검색했을 때 즐겨찾기 친구는 뷰에서 사라지므로 검색 상태에서 친구를 선택하면 즐겨찾기 내에 있는 동일한 친구도 체크해줘야함
        updateFriend(user, true)
    }

    fun removeUncheckedFriend(user: Item.User) {
        _pickedFriendsList.remove(user)
        _pickedFriends.value = _pickedFriendsList
        // 검색시 뷰에 없는 선택된 친구를 제거하고, 해당 친구가 뷰에 다시 노출될 때 화면 동기화
        // (FriendsListAdapter에서 단일 리스트를 사용하기 때문에 어뎁터에서는 검색에 의해 뷰에 안보이는 선택된 친구를 제거할 수 없음)
        updateFriend(user, false)
    }

    fun searchFriends(searchText: String) {
        if (searchText.isBlank()) {
            _isSearching.value = false
            _pickerList.value = totalList
        } else {
            _isSearching.value = true

            val filteredList =
                mutableListOf<Item>(Item.Header(this.context.getString(R.string.search_result)))
            filteredList.addAll(_friendsList.filter {
                PhonemeUtils.isMatch(it.nickname ?: "", searchText)
            })
            if (filteredList.size == 1) {
                filteredList.clear()
                filteredList.add(Item.EmptySearchResult)
            }
            _pickerList.value = filteredList
        }
    }

    private fun updateFriend(user: Item.User, isChecked: Boolean) {
        val indexOfFirst = totalList.indexOfFirst { it == user }
        val indexOfLast = totalList.indexOfLast { it == user }
        (totalList[indexOfFirst] as Item.User).isChecked = isChecked
        (totalList[indexOfLast] as Item.User).isChecked = isChecked
    }
}